from .exercise import Exercise
import os, json, random
from html import escape as escape_html
from .wrapper import sage

class Outcome():
    def __init__(self, title=None, slug=None, path=None, description=None, bank=None):
        self.title = title
        self.slug = slug
        self.relpath = path
        self.description = description
        self.bank = bank
    
    def abspath(self):
        return os.path.join(self.bank.abspath(),self.relpath)
    
    def full_title(self,max_length=None):
        ft = f"{self.slug}: {self.title}"
        if (max_length is not None) and (len(ft)>max_length):
            return ft[:max_length]+"…"
        else:
            return ft

    def template_filepath(self):
        return os.path.join(
            self.abspath(),
            "template.xml"
        )
    
    def template(self):
        with open(self.template_filepath()) as f:
            return f.read()

    def generator_path(self):
        return os.path.join(
            self.abspath(),
            "generator.sage"
        )

    def to_dict(self,regenerate=False):
        self.generate_exercises(regenerate)
        exs = self.exercises()
        return {
            "title": self.title,
            "slug": self.slug,
            "description": self.description,
            "template": self.template(),
            "exercises": [e.to_dict() for e in exs],
        }

    def preview_exercises(self):
        preview_json = os.path.join(self.build_path(),"preview.json")
        sage(self,preview_json,preview=True,images=True)
        with open(os.path.join(preview_json)) as f:
            data = json.load(f)['seeds']
        return [Exercise(d["data"],d["seed"],self) for d in data]

    def html_preview(self,pregenerated=False):
        if pregenerated:
            exs = random.sample(self.exercises(),1)
        else:
            exs = self.preview_exercises()
        html = "<h2>Preview:</h2>\n"
        for ex in exs:
            html += ex.html()
            html += "\n"
            html += "<h3>Data</h3>"
            html += "<pre>\n"
            html += escape_html(json.dumps(ex.to_dict(),indent=4))
            html += "</pre>\n"
            html += "\n"
            html += "<h3>SpaTeXt</h3>"
            html += "<pre>\n"
            html += escape_html(ex.spatext())
            html += "</pre>\n"
            html += "\n"
            html += "<h3>HTML</h3>"
            html += "<pre>\n"
            html += escape_html(ex.html())
            html += "</pre>\n"
            html += "<h3>LaTeX</h3>"
            html += "<pre>\n"
            html += escape_html(ex.latex())
            html += "</pre>\n"
            html += "<h3>PreTeXt</h3>"
            html += "<pre>\n"
            html += escape_html(ex.pretext())
            html += "</pre>\n"
        return html

    def build_path(self):
        p = os.path.join(self.bank.build_path(),self.slug,"generated")
        os.makedirs(p, exist_ok=True)
        return p
    
    def seeds_json_path(self):
        return os.path.join(self.build_path(),"seeds.json")

    def generate_exercises(self,regenerate=False,images=False):
        if not regenerate:
            try:
                self.load_exercises()
                return
            except RuntimeError:
                pass # generation is necessary
        sage(self,self.seeds_json_path(),preview=False,images=images)
        self.load_exercises(reload=True)


    def load_exercises(self,reload=False,strict=True):
        if not reload:
            try:
                self._exercises
            except AttributeError:
                pass # load is necessary
        try:
            with open(self.seeds_json_path()) as f:
                data = json.load(f)
            seed_list = data['seeds']
            self._exercises = [Exercise(d["data"],d["seed"],self) for d in seed_list]
            self._generated_on = data['generated_on']
        except FileNotFoundError as e:
            if strict:
                raise RuntimeError("Exercises must be generated before being loaded in strict mode.") from e
    
    def generated_on(self):
        try:
            return self._generated_on
        except AttributeError as e:
            return "(never generated)"
    
    def exercises(self,all=True,amount=300,randomized=False):
        try:
            exs = self._exercises
            if all:
                return exs
            if randomized:
                indices = sorted(random.sample(range(len(exs)),amount))
            else:
                indices = range(amount)
            return [exs[i] for i in indices]
        except AttributeError as e:
            raise RuntimeError("Exercises must be generated/loaded before being requested.") from e



from lxml import etree
from latex2mathml.converter import convert
import pystache
from .static import read_resource

def tex_to_mathml(tex):
    return etree.fromstring(convert(tex))

class Exercise:
    def __init__(self, data=None, seed=None, outcome=None):
        self.data = data
        self.seed = seed
        self.outcome = outcome

    def spatext_ele(self):
        renderer = pystache.Renderer()
        xml_string = renderer.render_path(self.outcome.template_filepath(),self.data)
        try:
            ele = etree.fromstring(bytes(xml_string, encoding='utf-8'))
        except etree.XMLSyntaxError as e:
            lined_xml = "\n".join([f"{i+1:04d}: {l}" for i,l in enumerate(xml_string.split("\n"))])
            e_text = str(e)+"\n"+lined_xml
            raise Exception(e_text) from e
        # remove comments
        etree.strip_tags(ele,etree.Comment)
        return ele

    def spatext(self):
        return str(
            etree.tostring(self.spatext_ele(), pretty_print=True), 
            encoding="UTF-8"
        )

    def html_ele(self,subset='all',consumer='basic'):
        transform = etree.XSLT(etree.fromstring(read_resource("html.xsl")))
        ele = transform(
            self.spatext_ele(),
            subset=f"'{subset}'",
            consumer=f"'{consumer}'",
            ).getroot()
        return ele

    def html(self,subset='all',consumer='basic'):
        return str(etree.tostring(self.html_ele(
            subset=subset,
            consumer=consumer
            ),pretty_print=True), 'utf-8')

    def pretext_ele(self,subset='all',consumer='basic'):
        transform = etree.XSLT(etree.fromstring(read_resource("pretext.xsl")))
        ele = transform(
            self.spatext_ele(),
            subset=f"'{subset}'",
            consumer=f"'{consumer}'",
            ).getroot()
        return ele

    def pretext(self,subset='all',consumer='basic'):
        return str(etree.tostring(self.pretext_ele(
            subset=subset,
            consumer=consumer
            ),pretty_print=True), 'utf-8')

    def latex(self):
        transform = etree.XSLT(etree.fromstring(read_resource("latex.xsl")))
        return str(transform(self.spatext_ele()))

    def to_dict(self):
        return {
            "seed": self.seed,
            "data": self.data,
        }



from IPython.display import display, Markdown, HTML
import ipywidgets as widgets
from .bank import Bank
from . import VERSION
from html import escape as escape_html

class modifiedOutput(widgets.Output):
    """
    https://github.com/jupyter-widgets/ipywidgets/issues/3208#issuecomment-1070836153
    """
    def __exit__(self, *args, **kwargs):
        super().__exit__(*args, **kwargs)

def run(bank=None):
    if bank is None:
        bank = Bank()
    menu_dropdown = widgets.Dropdown(
        options=[
            ('',''),
            ('Author/edit outcomes', 'outcome'),
            ('Manage bank', 'bank'),
        ],
        description='Menu:',
    )
    submenu = widgets.Output()
    menu_dropdown.observe(change_submenu(submenu,bank),names='value')
    display(Markdown(f"## {bank.title}"))
    display(HTML("<style>.stx-outtro{background-color:#ddd;}</style>"))
    display(menu_dropdown)
    display(submenu)
    display(Markdown("---"))
    display(Markdown(f"`CheckIt Dashboard v{VERSION}`"))

def change_submenu(submenu,bank):
    @submenu.capture(clear_output=True)
    def callback(value):
        if value['new'] == 'outcome':
            outcome_submenu(bank)
        elif value['new'] == 'bank':
            bank_submenu(bank)
    return callback

def outcome_submenu(bank): 
    options = [
        (f"{o.slug}: {o.title}",o) for o in bank.outcomes()
    ]
    outcomes_dropdown = widgets.Dropdown(options=options,
        description='Outcome:')
    preview_button = widgets.Button(description="Fresh preview")
    seed_button = widgets.Button(description="View random seed")
    build_button = widgets.Button(description="Generate seeds")
    images_button = widgets.Button(description="Gen seeds+graphics")
    description = modifiedOutput()
    generated = modifiedOutput()
    output = modifiedOutput()

    def reset(*args,only_generated=False):
        o = outcomes_dropdown.value
        if not only_generated:
            description.clear_output()
            with description:
                display(Markdown(f"**Description:** {escape_html(o.description)}"))
            output.clear_output()
        generated.clear_output()
        with generated:
            display(Markdown(f"*Last generated on:* `{o.generated_on()}`"))

    def preview(*args):
        o = outcomes_dropdown.value
        output.clear_output()
        with output:
            display(Markdown(f"*Generating fresh preview...*"))
        with output:
            preview = o.html_preview(pregenerated=False)
        with output:
            output.clear_output()
            display(HTML(preview))

    def seed(*args):
        o = outcomes_dropdown.value
        output.clear_output()
        with output:
            display(Markdown(f"*Selecting pregenerated seed...*"))
        with output:
            preview = o.html_preview(pregenerated=True)
        with output:
            output.clear_output()
            display(HTML(preview))

    def build(*args):
        o = outcomes_dropdown.value
        output.clear_output()
        with output:
            display(Markdown("Generating 1,000 seeds..."))
        with output:
            o.generate_exercises(regenerate=True)
        reset(only_generated=True)
        with output:
            display(Markdown("Done!"))

    def images(*args):
        o = outcomes_dropdown.value
        output.clear_output()
        with output:
            display(Markdown("Generating 1,000 seeds with graphics... (this can take some time)"))
        with output:
            o.generate_exercises(regenerate=True,images=True)
        reset(only_generated=True)
        with output:
            display(Markdown("Done!"))

    outcomes_dropdown.observe(reset,names="value")
    preview_button.on_click(preview)
    seed_button.on_click(seed)
    build_button.on_click(build)
    images_button.on_click(images)

    display(outcomes_dropdown)
    display(widgets.HBox([preview_button,seed_button,build_button,images_button]))
    display(description)
    display(generated)
    display(output)
    reset()

def bank_submenu(bank):
    build_label = widgets.Label(value="Build:")
    json_button = widgets.Button(description="Bank from cache")
    json_regen_button = widgets.Button(description="Regenerated bank")
    buttons = widgets.HBox([build_label,json_button,json_regen_button])
    output = modifiedOutput()
    generated = modifiedOutput()

    def reset(*args):
        output.clear_output()
        with output:
            display(Markdown(f"*Last built on:* `{bank.generated_on()}`"))

    def json(*args):
        generated.clear_output()
        with generated:
            display(Markdown("Building bank from pregenerated outcomes..."))
        with generated:
            bank.write_json()
        with generated:
            bank.build_viewer()
        reset()
        with generated:
            display(Markdown("Done!"))

    def json_regen(*args):
        generated.clear_output()
        with generated:
            display(Markdown("Regenerating all outcomes (no graphics) and building bank..."))
        with generated:
            bank.write_json(regenerate=True)
        with generated:
            bank.build_viewer()
        reset()
        with generated:
            display(Markdown("Done!"))

    json_button.on_click(json)
    json_regen_button.on_click(json_regen)

    display(buttons)
    display(output)
    display(generated)
    reset()



import importlib.resources
import subprocess, os, tempfile, shutil
from ..utils import working_directory

def sage(outcome,output_path,preview=True,images=False):
    """
    Wraps generator and builds to a seeds.json file at output_path
    """
    if preview:
        preview_s = "preview"
    else:
        preview_s = "build"
    if not os.path.isfile(outcome.generator_path()):
        raise FileNotFoundError(outcome.generator_path())
    with importlib.resources.path("checkit.wrapper", "wrapper.sage") as wrapper_path:
        with tempfile.TemporaryDirectory() as tmpdir:
            shutil.copyfile(wrapper_path,os.path.join(tmpdir,"wrapper.sage"))
            with working_directory(outcome.bank.abspath()):
                cmds = [
                    "sage",
                    os.path.join(tmpdir,"wrapper.sage"),
                    outcome.generator_path(),
                    output_path,
                    preview_s,
                ]
                if images:
                    cmds += ["images"]
                subprocess.run(cmds,check=True)





# This file was *autogenerated* from the file /home/user/checkit/dashboard/checkit/wrapper/wrapper.sage
from sage.all_cmdline import *   # import sage library

_sage_const_100000 = Integer(100000); _sage_const_999999 = Integer(999999); _sage_const_0 = Integer(0); _sage_const_1 = Integer(1); _sage_const_6 = Integer(6); _sage_const_3 = Integer(3); _sage_const_4 = Integer(4); _sage_const_1000 = Integer(1000); _sage_const_10 = Integer(10); _sage_const_1_000 = Integer(1_000); _sage_const_5 = Integer(5); _sage_const_2 = Integer(2)
import sys,json,os,datetime

# Library of helpful functions
class CheckIt:
    @staticmethod
    def vars(*latex_names, random_order=True):
        """
        Given one or more `latex_names` of strings, returns a tuple
        of Sage variables. `random_order` names them so that they appear
        in expressions in a random order.
        """
        stamp = randrange(_sage_const_100000 ,_sage_const_999999 )
        indices = list(range(len(latex_names)))
        if random_order:
            shuffle(indices)
        import string
        random_letter = choice(list(string.ascii_lowercase))
        return (var(f"{random_letter}_mi_var_{stamp}_{indices[i]}", latex_name=name) for i, name in enumerate(latex_names))

    @staticmethod
    def shuffled_equation(*terms):
        """
        Represents the equation sum(terms)==0, but with terms shuffled randomly
        to each side.
        """
        new_equation = (SR(_sage_const_0 )==_sage_const_0 )
        for term in terms:
            if choice([True,False]):
                new_equation += (SR(term)==_sage_const_0 )
            else:
                new_equation += (_sage_const_0 ==-SR(term))
        return new_equation*choice([-_sage_const_1 ,_sage_const_1 ])

    @staticmethod
    def shuffled_inequality(*terms,strict=True):
        """
        Represents the equation sum(terms)>0 or >=0, but with terms shuffled randomly
        to each side, and random direction of inequality
        """
        if choice([True,False]):
            if strict:
                new_equation = (SR(_sage_const_0 )>_sage_const_0 )
            else:
                new_equation = (SR(_sage_const_0 )>=_sage_const_0 )
            for term in terms:
                if choice([True,False]):
                    new_equation += (SR(term)==_sage_const_0 )
                else:
                    new_equation += (_sage_const_0 ==-SR(term))
        else:
            if strict:
                new_equation = (SR(_sage_const_0 )<_sage_const_0 )
            else:
                new_equation = (SR(_sage_const_0 )<=_sage_const_0 )
            for term in terms:
                if choice([True,False]):
                    new_equation += (-SR(term)==_sage_const_0 )
                else:
                    new_equation += (_sage_const_0 ==SR(term))
        return new_equation

    @staticmethod
    def latex_system_from_matrix(matrix, variables="x", alpha_mode=False, variable_list=[]):
        # Augment with zero vector if not already augmented
        if not matrix.subdivisions()[_sage_const_1 ]:
            matrix=matrix.augment(zero_vector(QQ, len(matrix.rows())), subdivide=true)
        num_vars = matrix.subdivisions()[_sage_const_1 ][_sage_const_0 ]
        # Start using requested variables
        system_vars = variable_list
        # Conveniently add xyzwv if requested
        if alpha_mode:
            system_vars += list(var("x y z w v"))
        # Finally fall back to x_n as needed
        system_vars += [var(f"{variables}_{n+_sage_const_1 }") for n in range(num_vars)]
        # Build matrix
        latex_output = "\\begin{matrix}\n"
        for row in matrix.rows():
            if row[_sage_const_0 ]!= _sage_const_0 :
                latex_output += latex(row[_sage_const_0 ]*system_vars[_sage_const_0 ])
                previous_terms = True
            else:
                previous_terms = False
            for n,cell in enumerate(row[_sage_const_1 :num_vars]):
                latex_output += " & "
                if cell < _sage_const_0 :
                    latex_output += " - "
                elif cell > _sage_const_0  and previous_terms:
                    latex_output += " + "
                latex_output += " & "
                if cell != _sage_const_0 :
                    latex_output += latex(cell.abs()*system_vars[n+_sage_const_1 ])
                if not previous_terms:
                    previous_terms = bool(cell!=_sage_const_0 )
            if not previous_terms:
                latex_output += " 0 "
            latex_output += " & = & "
            latex_output += latex(row[num_vars])
            latex_output += "\\\\\n"
        latex_output += "\\end{matrix}"
        return latex_output

    @staticmethod
    def latex_solution_set_from_matrix(matrix):
        # Augment with zero vector if not already augmented
        if not matrix.subdivisions()[_sage_const_1 ]:
            matrix=matrix.augment(zero_vector(QQ, len(matrix.rows())), subdivide=true)
        if (len(matrix.columns())-_sage_const_1 ) in matrix.pivots():
            return r" \{\} "
        solution_dimension = len(matrix.columns())-_sage_const_1 
        free_variables = list(var("a b c d e f g h i j"))
        kernel_basis=matrix.subdivision(_sage_const_0 ,_sage_const_0 ).right_kernel(basis='pivot').basis()
        span = sum([kernel_basis[i]*free_variables[i] for i in range(len(kernel_basis))])
        offset = zero_vector(QQ,solution_dimension)
        for row_index,col_index in enumerate(matrix.pivots()):
            offset[col_index] = matrix.rref().columns()[-_sage_const_1 ][row_index]
        rep = column_matrix(span+offset)
        predicate = ",".join([latex(a) for a in free_variables[:len(kernel_basis)]])
        return r" \left\{ " + latex(rep) + r" \,\middle|\, " + predicate + r" \in\mathbb R \right\} "

    @staticmethod
    def simple_random_matrix_of_rank(rank,rows=_sage_const_1 ,columns=_sage_const_1 ,augmented=False):
        # get extra rows and columns, at least zero
        extra_rows = max(_sage_const_0 ,rows-rank)
        extra_columns = max(_sage_const_0 ,columns-rank)
        # create matrix with terms between -5 and 5 inclusive, rank in every column, and integer entries RREF
        A = random_matrix(QQ,rank+extra_rows,rank,algorithm='echelonizable',rank=rank,upper_bound=_sage_const_6 )
        # randomly choose pivot indices where dependent columns are injected afterward
        inserts = [randrange(rank) for _ in range(extra_columns)]
        # pedagogically we want final column to be dependent at least half the time
        if extra_columns>_sage_const_0  and choice([True,False]):
            inserts[_sage_const_0 ]=rank-_sage_const_1 
        # we'll insert columns backwards to avoid messing up where to inject columns
        inserts.sort(reverse=True)
        # we won't repeat dependent columns
        inserted_columns = []
        for pivot in inserts:
            while True:
                # get random numbers for pivot rows
                rref_pivot_entries = [randrange(-_sage_const_3 ,_sage_const_4 ) for _ in range(pivot+_sage_const_1 )]
                # ensure at least one is nonzero
                rref_pivot_entries[randrange(pivot+_sage_const_1 )] = randrange(_sage_const_1 ,_sage_const_4 )*choice([-_sage_const_1 ,_sage_const_1 ])
                # create vector
                dependent_vector = sum([rref_pivot_entries[_]*A.column(_) for _ in range(pivot+_sage_const_1 )])
                if dependent_vector not in inserted_columns:
                    inserted_columns.append(dependent_vector)
                    A = matrix(A.columns()[:pivot+_sage_const_1 ]+[dependent_vector]+A.columns()[pivot+_sage_const_1 :]).transpose()
                    break
        if augmented:
            A.subdivide([],[columns-_sage_const_1 ])
        return A

# decorator to help authors avoid confusing .data() with .get_data() in a Generator
def provide_data(func):
    return lambda self: func(self.get_data())

# BaseGenerator class inherited by each outcome's Generator class to minimize boilerplate
class BaseGenerator:
    def __init__(self):
        self.__data = None
        self.__seed = None

    def data(self):
        return {}

    @provide_data
    def graphics(data):
        return None
    
    def roll_data(self,seed=None):
        if seed is None:
            set_random_seed()
            seed = randrange(_sage_const_1000 )
        self.__seed = seed
        set_random_seed(seed)
        self.__data = self.data()

    def get_data(self):
        data = self.__data
        data["__seed__"] = f"{self.__seed:04}"
        return self.__data

# converts SageMath objects into latexified strings
# note Python numbers are latexified into strings as well
def json_ready(obj):
    if isinstance(obj,str) or isinstance(obj,bool):
        return obj
    elif isinstance(obj,list):
        return [json_ready(item) for item in obj]
    elif isinstance(obj,dict):
        return {key:json_ready(obj[key]) for key in obj.keys()}
    else:
        return str(latex(obj))

# sage /path/to/wrapper.sage /path/to/generator.sage /path/to/output/seeds.json preview|build images?
if len(sys.argv) >= _sage_const_4 :
    # this script should be called from the root directory of the bank
    # so loads in the generator file work as intended
    generator_path = sys.argv[_sage_const_1 ]
    load(generator_path) # must provide Generator class extending BaseGenerator
    generator = Generator()

    # preview/build to specified JSON file
    if sys.argv[_sage_const_3 ].lower() == "preview":
        amount = _sage_const_10 
    else:
        amount = _sage_const_1_000 
    seeds = []
    for i in range(amount):
        if sys.argv[_sage_const_3 ].lower() == "preview":
            set_random_seed()
            seed_int = int(randrange(_sage_const_1_000 ))
        else:
            seed_int = int(i)
        gen_images = (len(sys.argv) >= _sage_const_5  and sys.argv[_sage_const_4 ]=="images")
        generator.roll_data(seed=seed_int)
        seed  = {"seed":seed_int,"data":json_ready(generator.get_data())}
        if gen_images:
            graphics = generator.graphics()
            if graphics is not None:
                directory = os.path.join(os.path.dirname(sys.argv[_sage_const_2 ]))
                if not os.path.exists(directory):
                    os.makedirs(directory)
                for filename in graphics:
                    seed_path = os.path.join(directory,f"{seed_int:04}")
                    if not os.path.exists(seed_path):
                        os.makedirs(seed_path)
                    graphics[filename].save(os.path.join(seed_path,f"{filename}.png"))
        seeds.append(seed)
    data = {
        "seeds": seeds,
        "generated_on": datetime.datetime.now(datetime.timezone.utc).isoformat(),
    }
    os.makedirs(os.path.dirname(sys.argv[_sage_const_2 ]), exist_ok=True)
    with open(os.path.join(sys.argv[_sage_const_2 ]), 'w') as f:
        json.dump(data, f)




import importlib.resources

def read_resource(resource_name):
    return importlib.resources.read_text("checkit.static", resource_name)

def open_resource(resource_name):
    return importlib.resources.open_binary("checkit.static", resource_name)


import click
import os
from . import static, VERSION

@click.command()
@click.argument('directory', default='new-checkit-bank')
def main(directory):
    """
    Generates boilerplate for using the CheckIt Dashboard
    to author a new bank.
    """
    # create directories
    try:
        os.makedirs(directory)
    except FileExistsError:
        print(f"Error creating new bank: directory `{directory}` already exists")
        return
    # copy sample outcome template/generator
    example_outcome_dir = os.path.join(directory,'outcomes','EX1')
    os.makedirs(example_outcome_dir)
    for filename in ["template.xml","generator.sage"]:
        with open(os.path.join(example_outcome_dir,filename),"w") as f:
            f.write(static.read_resource(filename))
    # copy dashboard notebook, bank manifest, README
    for filename in ["dashboard.ipynb","bank.xml","README.md"]:
        with open(os.path.join(directory,filename),"w") as f:
            f.write(static.read_resource(filename))
    # copy gitignore
    with open(os.path.join(directory,".gitignore"),"w") as f:
        f.write(static.read_resource("gitignore.txt"))
    # generate requirements.txt
    with open(os.path.join(directory,"requirements.txt"),"w") as f:
        f.write(f"checkit-dashboard == {VERSION}")
    print(f"Successfully created new CheckIt bank in `{directory}`")

if __name__ == "__main__":
    main()



import os
from contextlib import contextmanager

@contextmanager
def working_directory(path):
    """
    Temporarily change the current working directory.
    Usage:
    with working_directory(path):
        do_things()   # working in the given path
    do_other_things() # back to original path
    """
    current_directory=os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(current_directory)


from lxml import etree
import os, json, datetime, zipfile, shutil, glob
from . import static
from .outcome import Outcome
from .xml import CHECKIT_NS

class Bank():
    def __init__(self, path="."):
        # read manifest for bank
        self._abspath = os.path.abspath(path)
        xml = etree.parse(os.path.join(self.abspath(),"bank.xml")).getroot()
        if xml.get("version") != "0.2":
            raise Exception("ERROR: Bank configuration doesn't match CheckIt version 0.2")
        self.title = xml.find(f"{CHECKIT_NS}title").text
        self.slug = xml.find(f"{CHECKIT_NS}slug").text
        self.url = xml.find(f"{CHECKIT_NS}url").text
        # create each outcome
        self._outcomes = [
            Outcome(
                ele.find(f"{CHECKIT_NS}title").text,
                ele.find(f"{CHECKIT_NS}slug").text,
                ele.find(f"{CHECKIT_NS}path").text,
                ele.find(f"{CHECKIT_NS}description").text,
                self,
            )
            for ele in xml.find(f"{CHECKIT_NS}outcomes").iter(f"{CHECKIT_NS}outcome")
        ]
        for o in self._outcomes:
            o.load_exercises(strict=False)
    
    def abspath(self):
        return self._abspath
    
    def outcomes(self):
        return self._outcomes
    
    def generate_exercises(self,regenerate=False,images=False):
        for o in self.outcomes():
            o.generate_exercises(regenerate=regenerate,images=images)

    def build_path(self):
        p = os.path.join(self.abspath(),"assets")
        os.makedirs(p, exist_ok=True)
        return p

    def to_dict(self,regenerate=False):
        olist = [o.to_dict(regenerate=regenerate) for o in self.outcomes()]
        return {
            "title": self.title,
            "slug": self.slug,
            "url": self.url,
            "generated_on": datetime.datetime.now(datetime.timezone.utc).isoformat(),
            "outcomes": olist,
        }

    def write_json(self,regenerate=False):
        build_path = os.path.join(self.build_path(),f"bank.json")
        with open(build_path,'w') as f:
            json.dump(self.to_dict(regenerate=regenerate),f)

    def build_viewer(self):
        build_path = os.path.join(self.abspath(),"docs")
        if os.path.exists(build_path) and os.path.isdir(build_path):
            shutil.rmtree(build_path)
        os.makedirs(build_path)
        archive = zipfile.ZipFile(static.open_resource("viewer.zip"))
        archive.extractall(build_path)
        # copy assets
        shutil.copytree(self.build_path(),os.path.join("docs","assets"), dirs_exist_ok=True)

    def generated_on(self):
        try:
            with open(os.path.join(self.build_path(),f"bank.json"),'r') as f:
                return json.load(f)["generated_on"]
        except:
            return "(never generated)"



from setuptools import setup

# Metadata goes in setup.cfg. These are here for GitHub's dependency graph.
setup(
    name="checkit-dashboard",
    install_requires=[
        "ipywidgets",
        "lxml",
        "latex2mathml",
        "pystache",
        "click",
    ],
)
